'use strict';

var _ramda = require('ramda');

var _ramda2 = _interopRequireDefault(_ramda);

var _utils = require('../../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * this file had been forked from https://github.com/dependents/node-dependency-tree
 */

var precinct = require('../precinct');
var fs = require('fs');
var cabinet = require('../filing-cabinet');
var debug = require('debug')('tree');
var Config = require('./Config');

/**
 * Recursively find all dependencies (avoiding circular) traversing the entire dependency tree
 * and returns a flat list of all unique, visited nodes
 *
 * @param {Object} options
 * @param {String} options.filename - The path of the module whose tree to traverse
 * @param {String} options.directory - The directory containing all JS files
 * @param {String} [options.requireConfig] - The path to a requirejs config
 * @param {String} [options.webpackConfig] - The path to a webpack config
 * @param {Object} [options.visited] - Cache of visited, absolutely pathed files that should not be reprocessed.
 *                             Format is a filename -> tree as list lookup table
 * @param {Array} [options.nonExistent] - List of partials that do not exist
 * @return {Object}
 */
module.exports = function (options) {
  var config = new Config(options);

  if (!fs.existsSync(config.filename)) {
    debug(`file ${config.filename} does not exist`);
    return {};
  }

  return traverse(config);
};

/**
 * Executes a post-order depth first search on the dependency tree and returns a
 * list of absolute file paths. The order of files in the list will be the
 * proper concatenation order for bundling.
 *
 * In other words, for any file in the list, all of that file's dependencies (direct or indirect) will appear at
 * lower indices in the list. The root (entry point) file will therefore appear last.
 *
 * The list will not contain duplicates.
 *
 * Params are those of module.exports
 */
module.exports.toList = function (options) {
  options.isListForm = true;

  return module.exports(options);
};

/**
 * Returns the list of dependencies for the given filename
 *
 * Protected for testing
 *
 * @param  {Config} config
 * @return {Array}
 */
module.exports._getDependencies = function (config) {
  var dependenciesRaw = void 0; // from some detectives it comes as an array, from some it is an object
  var dependencies = void 0; // always an array
  var precinctOptions = config.detectiveConfig;
  precinctOptions.includeCore = false;

  try {
    dependenciesRaw = precinct.paperwork(config.filename, precinctOptions);
  } catch (e) {
    debug(`error getting dependencies: ${e.message}`);
    debug(e.stack);
    e.code = 'PARSING_ERROR';
    config.errors[config.filename] = e;
    dependenciesRaw = [];
  }
  dependencies = _ramda2.default.is(Object, dependenciesRaw) && !Array.isArray(dependenciesRaw) ? Object.keys(dependenciesRaw) : dependenciesRaw;
  var isDependenciesArray = Array.isArray(dependenciesRaw);
  debug(`extracted ${dependencies.length} dependencies: `, dependencies);

  var resolvedDependencies = [];
  var pathMapDependencies = [];
  var pathMapFile = { file: config.filename };

  for (var i = 0, l = dependencies.length; i < l; i++) {
    var dependency = dependencies[i];
    var cabinetParams = {
      partial: dependency,
      filename: config.filename,
      directory: config.directory,
      ast: precinct.ast,
      config: config.requireConfig,
      webpackConfig: config.webpackConfig,
      resolveConfig: config.resolveConfig
    };
    if (!isDependenciesArray && dependenciesRaw[dependency].isScript !== undefined) {
      // used for vue
      cabinetParams.isScript = dependenciesRaw[dependency].isScript;
    }
    var result = void 0;
    try {
      result = cabinet(cabinetParams);
    } catch (err) {
      debug(`error resolving dependencies: ${err.message}`);
      debug(err.stack);
      err.code = 'RESOLVE_ERROR';
      throw err;
    }

    if (!result) {
      debug(`skipping an empty filepath resolution for partial: ${dependency}`);
      if (config.nonExistent[config.filename]) {
        config.nonExistent[config.filename].push(dependency);
      } else {
        config.nonExistent[config.filename] = [dependency];
      }
      continue;
    }

    var exists = fs.existsSync(result);

    if (!exists) {
      if (config.nonExistent[config.filename]) {
        config.nonExistent[config.filename].push(dependency);
      } else {
        config.nonExistent[config.filename] = [dependency];
      }
      debug(`skipping non-empty but non-existent resolution: ${result} for partial: ${dependency}`);
      continue;
    }
    var pathMap = { importSource: dependency, resolvedDep: result };
    if (!isDependenciesArray && dependenciesRaw[dependency].importSpecifiers) {
      pathMap.importSpecifiers = dependenciesRaw[dependency].importSpecifiers;
    }
    if (!(0, _utils.isRelativeImport)(dependency) && config.resolveConfig) {
      // is includes also packages, which actually don't use customResolve, however, they will be
      // filtered out later.
      pathMap.isCustomResolveUsed = true;
    }

    pathMapDependencies.push(pathMap);

    resolvedDependencies.push(result);
  }
  pathMapFile.dependencies = pathMapDependencies;
  config.pathMap.push(pathMapFile);

  return resolvedDependencies;
};

/**
 * the traverse is not a recursive function anymore, it has been changed to be iterative to fix
 * some performance issues.
 * @todo: we have some redundancy here with the `tree` and pathMap`. the `tree` has only the dependencies,
 * `pathMap` has the dependencies and some more info, such as importSpecifiers. we should use only
 * pathMap and get rid of tree.
 */
function traverse(config) {
  var tree = [];
  var stack = [config.filename];

  var _loop = function _loop() {
    var dependency = stack.pop();
    debug(`traversing ${dependency}`);
    if (!config.visited[dependency]) {
      var localConfig = config.clone();
      localConfig.filename = dependency;
      var dependencies = module.exports._getDependencies(localConfig);
      if (config.filter) {
        debug('using filter function to filter out dependencies');
        debug(`number of dependencies before filtering: ${dependencies.length}`);
        dependencies = dependencies.filter(function (filePath) {
          return localConfig.filter(filePath, localConfig.filename);
        });
        debug(`number of dependencies after filtering: ${dependencies.length}`);
      }
      debug('cabinet-resolved all dependencies: ', dependencies);
      tree[dependency] = dependencies;
      var filePathMap = config.pathMap.find(function (pathMapEntry) {
        return pathMapEntry.file === dependency;
      });
      if (!filePathMap) throw new Error(`file ${dependency} is missing from PathMap`);
      config.visited[dependency] = filePathMap;
      dependencies.forEach(function (d) {
        return stack.push(d);
      });
    } else {
      (function () {
        debug(`already visited ${dependency}. Will try to find it and its dependencies in the cache`);
        var dependenciesStack = [dependency];
        while (dependenciesStack.length) {
          var dep = dependenciesStack.pop();
          if (!config.visited[dep]) {
            debug(`unable to find ${dep} in the cache, it was probably filtered before`);
            continue;
          }
          debug(`found ${dep} in the cache`);
          var _dependencies = config.visited[dep].dependencies.map(function (d) {
            return d.resolvedDep;
          });
          tree[dep] = _dependencies;
          config.pathMap.push(config.visited[dep]);
          _dependencies.forEach(function (d) {
            if (!tree[d]) dependenciesStack.push(d);
          });
        }
      })();
    }
  };

  while (stack.length) {
    _loop();
  }

  return tree;
}
//# sourceMappingURL=index.js.map